package org.livee.data.query.resolve;
/*
 * Copyright (c) 2007 Hideya Ochiai,  All rights reserved.
 * 
 * Permission of redistribution and use in source and binary forms, 
 * with or without modification, are granted, free of charge, to any person 
 * obtaining the copy of this software under the following conditions:
 * 
 *  1. Any copies of this source code must include the above copyright notice,
 *  this permission notice and the following statement without modification 
 *  except possible additions of other copyright notices. 
 * 
 *  2. Redistributions of the binary code must involve the copy of the above 
 *  copyright notice, this permission notice and the following statement 
 *  in documents and/or materials provided with the distribution.  
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

import org.livee.data.exception.NoServerResolvedLiveEException;
import org.livee.data.exception.LiveEException;
import org.livee.data.expr.AbstractData;
import org.livee.data.query.QueryProcessCommand;
import org.livee.data.db.DatabaseAdapter;
import org.livee.data.query.DataFactory;

import org.livee.data.soap.client.Admin200703_PortType;
import org.livee.data.soap.client.Admin200703Service;
import org.livee.data.soap.client.Admin200703ServiceLocator;

/**
 * 
 * @author Hideya Ochiai, The University of Tokyo
 */
public class QueryResolver {

	public class QueryResolverCache extends Thread {
		
		public class CacheEntry {
			public AbstractData queryResponse;
			public long cachedTime;
			public CacheEntry(AbstractData qr,long time){
				queryResponse=qr;
				cachedTime=time;
			}
		}
		
		java.util.Map m_CacheMap;
		int m_MaxCacheSize;
		long m_CacheValidityTime;
		
		public QueryResolverCache(){
			m_CacheMap=new java.util.Hashtable();

			org.livee.data.tools.ConfigManager config=org.livee.data.tools.ConfigManager.getInstance();
			int validityTime=config.getIntParameter("/c:liveeConfiguration/c:search/c:queryResolver/c:cacheValidityTime/text()");
			int cacheSize=config.getIntParameter("/c:liveeConfiguration/c:search/c:queryResolver/c:cacheSize/text()");
			if(validityTime>0){
				m_CacheValidityTime=validityTime*1000;				
			}else{
				m_CacheValidityTime=600000; // recommended default 10min
			}
			if(cacheSize>0){
				m_MaxCacheSize=cacheSize;
			}else{
				m_MaxCacheSize=100; // recommended default 100
			}
			start();
		}
		
		public AbstractData getCachedQueryResponse(String query){
			Object o=m_CacheMap.get(query);
			if(o!=null){
				CacheEntry entry=(CacheEntry)o;
				//System.out.println("GET SUCCEEDED: "+query);
				return entry.queryResponse;
			}
			return null;
		}
		
		public void setQueryResponse(String query, AbstractData queryResponse, long current){
			CacheEntry entry=new CacheEntry(queryResponse,current);
			if(m_CacheMap.size()<m_MaxCacheSize){
				//System.out.println("PUT SUCCEEDED: "+query);
				synchronized(m_CacheMap){
					m_CacheMap.put(query, entry);
				}
			}
		}
		
		public void run(){
			// Cache removal 
			while(true){
				java.util.ArrayList keyset=new java.util.ArrayList();
				synchronized(m_CacheMap){
					keyset.addAll(m_CacheMap.keySet());
				}
				for(int i=0;i<keyset.size();i++){
					long current=System.currentTimeMillis();
					String key=(String)keyset.get(i);
					synchronized(m_CacheMap){
						CacheEntry entry=(CacheEntry)m_CacheMap.get(key);
						if((current-m_CacheValidityTime)>entry.cachedTime || current<entry.cachedTime){
							m_CacheMap.remove(key);
						}
					}
					try{
						Thread.sleep(300);
					}
					catch(Exception e){}
				}
			
				try{
					Thread.sleep(10000);
				}catch(Exception e){
					
				}	
			}
		}
	}
	
	static QueryResolver m_Singleton=null;
	QueryResolverCache m_Cache=null;
	
	protected QueryResolver(){
		m_Singleton=this;
		m_Cache=new QueryResolverCache();
	}
	
	public static QueryResolver getInstance(){
		if(m_Singleton==null){
			new QueryResolver();
		}
		return m_Singleton;
	}
	
	public AbstractData resolve(AbstractData query,String query_str) throws LiveEException {

		AbstractData queryResponse=null;
		
		// Check cache
		queryResponse=m_Cache.getCachedQueryResponse(query_str);
		if(queryResponse!=null){
			return queryResponse;
		}
		
		java.util.Map adminServerMap=new java.util.Hashtable();
		
		// Local Transaction
		QueryProcessCommand command=new QueryProcessCommand(query);
		DatabaseAdapter db=DatabaseAdapter.getInstance();
		
		try{
			db.postCommand(command);
			if(command.succeeded()){
				queryResponse=command.getQueryResponse();
			}
		}
		catch(LiveEException e){
			throw e;
		}
		catch(Exception e){
			e.printStackTrace();
		}
		
		if(queryResponse!=null){
			evalResponse(query_str,queryResponse,adminServerMap,new java.util.Hashtable());
		}
		
		if(adminServerMap.size()>0){
			queryResponse=DataFactory.getInstance().create("queryResponse");
			java.util.Iterator itr=adminServerMap.values().iterator();
			while(itr.hasNext()){
				queryResponse.append((AbstractData)itr.next());
			}
		}else{
			throw new NoServerResolvedLiveEException(query);
		}
		
		m_Cache.setQueryResponse(query_str, queryResponse, System.currentTimeMillis());
		return queryResponse;
	}
	
	
	protected void evalResponse(String query,AbstractData queryResponse,java.util.Map adminServerMap,java.util.Map triedServerMap){
		
		if(queryResponse==null){
			return ;
		}
		
		java.util.Iterator itr=queryResponse.getChildren().iterator();
		while(itr.hasNext()){
			AbstractData data=(AbstractData)itr.next();
			if(data==null){
				continue ;
			}
			if(data.getDataType().equals("AdminSite")){
				String name=data.getAttribute("name");
				if(!adminServerMap.containsKey(name)){
					adminServerMap.put(name,data);
				}
			}else if(data.getDataType().equals("RedirectTo")){
				java.util.Iterator itr2=data.getChildren().iterator();
				while(itr2.hasNext()){
					AbstractData adminSite=(AbstractData)itr2.next();
					String name=adminSite.getAttribute("name");
					if(!triedServerMap.containsKey(name)){
						triedServerMap.put(name, "accessed");
						// 	Request to the server
						AbstractData response=queryRequest(name,query,adminSite);
						if(response!=null){
							evalResponse(query,response,adminServerMap,triedServerMap);
						}
					}
				}
			}
		}
	}
	
	
	private AbstractData queryRequest(String name,String query,AbstractData adminSite){
		
		if(adminSite==null){
			return null;
		}
		
		java.util.Iterator itr=adminSite.getChildren().iterator();
		while(itr.hasNext()){
			AbstractData adminServer=(AbstractData)itr.next();
			String service=adminServer.getAttribute("service");
			String url=adminServer.getAttribute("url");
			String linkStatus=adminServer.getAttribute("linkStatus");
			
			if(linkStatus==null || (linkStatus!=null && linkStatus.equals("up"))){
				if(service.equals("Admin200703")){
					try{
						Admin200703Service admin=new Admin200703ServiceLocator();
						Admin200703_PortType srv=admin.getAdmin200703(new java.net.URL(url));
						
						//System.out.println("REQUEST:  "+query);
						String response=srv.search(query);
						//System.out.println("RESPONSE: "+response);
						
						if(response!=null){
							org.w3c.dom.Document doc=org.livee.util.Xml.parseXml(response);
							org.w3c.dom.Element xml=doc.getDocumentElement();
							org.livee.data.expr.AbstractParser queryResponse_parser=org.livee.data.query.ParserFactory.getInstance().get("queryResponse");
							return queryResponse_parser.parse(xml);
						}
					}
					catch(Exception e){
					//	e.printStackTrace();
					}
				}
			}
		}
				
		return null;
	}
}
